/*
 * Copyright 2018-2019 ON Semiconductor
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <linux/kernel.h>
#include <linux/device.h>
#include <linux/module.h>
#include <linux/i2c.h>
#include <linux/of_gpio.h>
#include <linux/of_device.h>
#include <linux/delay.h>
#include <linux/firmware.h>
#include <linux/debugfs.h>
#include <linux/version.h>
#include <linux/err.h>
#include <linux/input.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/mutex.h>
#include "fusb251.h"

#define I2C_RETRIES 2
#define I2C_RETRY_DELAY 5 /* ms */

bool is_usb_attached;
struct fusb251 *g_fusb251;

enum FUSB251_STATE_TYPE {
	TYPE_DRY = 0,
	TYPE_MOISTURE = 1,
};

static int fusb251_read_data(struct fusb251 *fusb251, unsigned char reg,
			int len, unsigned char value[])
{
	struct i2c_client *client;
	int err;
	int tries = 0;
	int error = 0;
	struct i2c_msg msgs[] = {
		{
			.flags = 0,
			.len = 1,
			.buf = &reg,
		}, {
			.flags = I2C_M_RD,
			.len = len,
			.buf = value,
		},
	};

	if (!fusb251) {
		pr_err("[fusb251] read data: No device is available\n");
		return -1;
	}

	mutex_lock(&fusb251->fusb251_i2c_mutex);

	client = fusb251->i2c;
	msgs[0].addr = client->addr;
	msgs[1].addr = client->addr;

	do {
		err = i2c_transfer(client->adapter, msgs, ARRAY_SIZE(msgs));
		if (err != ARRAY_SIZE(msgs))
			msleep_interruptible(I2C_RETRY_DELAY);
	} while ((err != ARRAY_SIZE(msgs)) && (++tries < I2C_RETRIES));

	if (err != ARRAY_SIZE(msgs)) {
		pr_err("[fusb251] i2c_read error, %d\n", err);
		error = -1;
	}

	mutex_unlock(&fusb251->fusb251_i2c_mutex);

	return error;
}

static int fusb251_write_data(struct fusb251 *fusb251,
			char *writebuf, int writelen)
{
	int ret = 0;
	struct i2c_client *client;

	if (fusb251->i2c == NULL) {
		pr_err("[fusb251] write data: No device is available\n");
		return -1;
	}

	mutex_lock(&fusb251->fusb251_i2c_mutex);

	client = fusb251->i2c;
	if (writelen > 0) {
		struct i2c_msg msgs[] = {
			{
				.addr = client->addr,
				.flags = 0,
				.len = writelen,
				.buf = writebuf,
			},
		};
		ret = i2c_transfer(client->adapter, msgs, 1);
		if (ret < 0)
			pr_err("[fusb251] i2c_write error, ret=%d\n", ret);
	}

	mutex_unlock(&fusb251->fusb251_i2c_mutex);

	return ret;
}

static int fusb251_i2c_write_reg(struct fusb251 *fusb251,
			unsigned char regaddr, unsigned char regvalue)
{
	unsigned char buf[2] = {0};

	if (!fusb251) {
		pr_err("[fusb251] write reg: No device is available\n");
		return -1;
	}

	buf[0] = regaddr;
	buf[1] = regvalue;

	return fusb251_write_data(fusb251, buf, sizeof(buf));
}

static void fusb251_read_interface(struct fusb251 *fusb251,
				unsigned char reg, unsigned char mask,
				unsigned char shift, unsigned char *value)
{
	unsigned char reg_val = 0;

	fusb251_read_data(fusb251, reg, 1, &reg_val);

	reg_val &= (mask << shift);
	*value = (reg_val >> shift);
}

static void fusb251_config_interface(struct fusb251 *fusb251,
				unsigned char reg, unsigned char mask,
				unsigned char shift, int value)
{
	unsigned char reg_val = 0;

	fusb251_read_data(fusb251, reg, 1, &reg_val);

	reg_val &= ~(mask << shift);
	reg_val |= (value << shift);

	fusb251_i2c_write_reg(fusb251, reg, reg_val);
}

static int fusb251_resistance_conver_reg(int value)
{
	int i = 0;

	for (i = 0; i < ARRAY_SIZE(mos_det_R_table); i++) {
		if (mos_det_R_table[i] == value)
			return i;
	}

	return -1;
}

static void fusb251_init(struct fusb251 *fusb251);
static ssize_t reset_store(struct device *dev,
				struct device_attribute *attr,
				const char *buff, size_t size)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	fusb251_i2c_write_reg(fusb251, FUSB251_RESET, RESET_DEVICE);
	mdelay(10);
	fusb251_init(fusb251);
	fusb251_i2c_write_reg(fusb251, FUSB251_CONTROL,
			CTL_EN_CC_DET | CTL_EN_SBU_DET | CTL_EN_DRY_DET);

	switch_set_state(&fusb251->st_switch, TYPE_DRY);

	pr_info("[fusb251] reset device and ld_switch\n");

	return size;
}

static ssize_t dry_det_store(struct device *dev,
				struct device_attribute *attr,
				const char *buff, size_t size)
{
	int temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	if (kstrtouint(buff, 0, &temp))
		return -EINVAL;

	if (temp) {
		fusb251_config_interface(fusb251, FUSB251_CONTROL,
				EN_DRY_DET_MASK, EN_DRY_DET_SHIFT, true);
		pr_info("[fusb251] enable dry detection\n");
	} else {
		fusb251_config_interface(fusb251, FUSB251_CONTROL,
				EN_DRY_DET_MASK, EN_DRY_DET_SHIFT, false);
		pr_info("[fusb251] disable dry detection\n");
	}

	return size;
}

static ssize_t dry_det_show(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	unsigned char temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	fusb251_read_interface(fusb251, FUSB251_CONTROL,
			EN_DRY_DET_MASK, EN_DRY_DET_SHIFT, &temp);

	return snprintf(buf, PAGE_SIZE, "%s\n", temp ? "enable" : "disable");
}

static ssize_t sbu_det_store(struct device *dev,
				struct device_attribute *attr,
				const char *buff, size_t size)
{
	int temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	if (kstrtouint(buff, 0, &temp))
		return -EINVAL;

	if (temp) {
		fusb251_config_interface(fusb251, FUSB251_CONTROL,
				EN_SBU_DET_MASK, EN_SBU_DET_SHIFT, true);
		pr_info("[fusb251] enable sbu detection\n");
	} else {
		fusb251_config_interface(fusb251, FUSB251_CONTROL,
				EN_SBU_DET_MASK, EN_SBU_DET_SHIFT, false);
		pr_info("[fusb251] disable sbu detection\n");
	}

	return size;
}

static ssize_t sbu_det_show(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	unsigned char temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	fusb251_read_interface(fusb251, FUSB251_CONTROL,
			EN_SBU_DET_MASK, EN_SBU_DET_SHIFT, &temp);

	return snprintf(buf, PAGE_SIZE, "%s\n", temp ? "enable" : "disable");
}

static ssize_t auto_sbu_det_store(struct device *dev,
				struct device_attribute *attr,
				const char *buff, size_t size)
{
	int temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	if (kstrtouint(buff, 0, &temp))
		return -EINVAL;

	if (temp) {
		fusb251_config_interface(fusb251, FUSB251_CONTROL,
				EN_AUTO_SBU_DET_MASK, EN_AUTO_SBU_DET_SHIFT, true);
		pr_info("[fusb251] enable auto sbu detection\n");
	} else {
		fusb251_config_interface(fusb251, FUSB251_CONTROL,
				EN_AUTO_SBU_DET_MASK, EN_AUTO_SBU_DET_SHIFT, false);
		pr_info("[fusb251] disable auto sbu detection\n");
	}

	return size;
}

static ssize_t auto_sbu_det_show(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	unsigned char temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	fusb251_read_interface(fusb251, FUSB251_CONTROL,
			EN_AUTO_SBU_DET_MASK, EN_AUTO_SBU_DET_SHIFT, &temp);

	return snprintf(buf, PAGE_SIZE, "%s\n", temp ? "enable" : "disable");
}

static ssize_t cc_det_store(struct device *dev,
				struct device_attribute *attr,
				const char *buff, size_t size)
{
	int temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	if (kstrtouint(buff, 0, &temp))
		return -EINVAL;

	if (temp) {
		fusb251_config_interface(fusb251, FUSB251_CONTROL,
				EN_CC_DET_MASK, EN_CC_DET_SHIFT, true);
		pr_info("[fusb251] enable cc detection\n");
	} else {
		fusb251_config_interface(fusb251, FUSB251_CONTROL,
				EN_CC_DET_MASK, EN_CC_DET_SHIFT, false);
		pr_info("[fusb251] disable cc detection\n");
	}

	return size;
}

static ssize_t cc_det_show(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	unsigned char temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	fusb251_read_interface(fusb251, FUSB251_CONTROL,
			EN_CC_DET_MASK, EN_CC_DET_SHIFT, &temp);

	return snprintf(buf, PAGE_SIZE, "%s\n", temp ? "enable" : "disable");
}

static ssize_t status_show(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	unsigned char temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	fusb251_read_data(fusb251, FUSB251_STATUS, 1, &temp);

	return snprintf(buf, PAGE_SIZE, "0x%02x\n", temp);
}

static ssize_t moi_status_show(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	unsigned char temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	fusb251_read_data(fusb251, FUSB251_MOISTATUS, 1, &temp);

	return snprintf(buf, PAGE_SIZE, "0x%02x\n", temp);
}

static ssize_t switch_cc_store(struct device *dev,
				struct device_attribute *attr,
				const char *buff, size_t size)
{
	int temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	if (kstrtouint(buff, 0, &temp))
		return -EINVAL;

	if (temp) {
		fusb251_config_interface(fusb251, FUSB251_SWITCHCONTROL,
				SWITCH_CC_MASK, SWITCH_CC_SHIFT, true);
		pr_info("[fusb251] close CC switch\n");
	} else {
		fusb251_config_interface(fusb251, FUSB251_SWITCHCONTROL,
				SWITCH_CC_MASK, SWITCH_CC_SHIFT, false);
		pr_info("[fusb251] open CC switch\n");
	}

	return size;
}

static ssize_t switch_cc_show(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	unsigned char temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	fusb251_read_interface(fusb251, FUSB251_SWITCHCONTROL,
			SWITCH_CC_MASK, SWITCH_CC_SHIFT, &temp);

	return snprintf(buf, PAGE_SIZE, "%s\n", temp ? "close" : "open");
}

static ssize_t switch_sbu_store(struct device *dev,
				struct device_attribute *attr,
				const char *buff, size_t size)
{
	int temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	if (kstrtouint(buff, 0, &temp))
		return -EINVAL;

	if (temp == 0x01) {
		fusb251_config_interface(fusb251, FUSB251_SWITCHCONTROL,
				SWITCH_SBU_MASK, SWITCH_SBU_SHIFT, true);
		pr_info("[fusb251] close sbu switch\n");
		return size;
	} else if (temp == 0x00) {
		fusb251_config_interface(fusb251, FUSB251_SWITCHCONTROL,
				SWITCH_SBU_MASK, SWITCH_SBU_SHIFT, false);
		pr_info("[fusb251] open sbu switch\n");
		return size;
	} else {
		pr_info("[fusb251] input error\n");
		return -EINVAL;
	}
}

static ssize_t switch_sbu_show(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	unsigned char temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	fusb251_read_interface(fusb251, FUSB251_SWITCHCONTROL,
			SWITCH_SBU_MASK, SWITCH_SBU_SHIFT, &temp);

	return snprintf(buf, PAGE_SIZE, "%s\n", temp ? "close" : "open");
}

static ssize_t threshold1_sbu_store(struct device *dev,
				struct device_attribute *attr,
				const char *buff, size_t size)
{
	int temp = 0;
	unsigned char ret = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	if (kstrtouint(buff, 0, &temp))
		return -EINVAL;

	if (temp >= 0 && temp <= MAX_RESISTANCE) {
		ret = fusb251_resistance_conver_reg(temp);
		if (ret >= 0) {
			fusb251_config_interface(fusb251, FUSB251_THRESHOLD1,
					SBU_MOS_R_DET_MASK, SBU_MOS_R_DET_SHIFT, ret);

			pr_info("[fusb251] set threshold1_sbu to %dk\n", temp);

			return size;
		}
	}

	return -EINVAL;
}

static ssize_t threshold1_sbu_show(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	unsigned char temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	fusb251_read_interface(fusb251, FUSB251_THRESHOLD1,
			SBU_MOS_R_DET_MASK, SBU_MOS_R_DET_SHIFT, &temp);

	return snprintf(buf, PAGE_SIZE, "%d\n", mos_det_R_table[temp]);
}

static ssize_t threshold1_cc_store(struct device *dev,
				struct device_attribute *attr,
				const char *buff, size_t size)
{
	int temp = 0;
	unsigned char ret = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	if (kstrtouint(buff, 0, &temp))
		return -EINVAL;

	if (temp >= 0 && temp <= MAX_RESISTANCE) {
		ret = fusb251_resistance_conver_reg(temp);
		if (ret >= 0) {
			fusb251_config_interface(fusb251, FUSB251_THRESHOLD1,
					CC_MOS_R_DET_MASK, CC_MOS_R_DET_SHIFT, ret);

			pr_info("[fusb251] set threshold1_cc to %dk\n", temp);

			return size;
		}
	}

	return -EINVAL;
}

static ssize_t threshold1_cc_show(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	unsigned char temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	fusb251_read_interface(fusb251, FUSB251_THRESHOLD1,
			CC_MOS_R_DET_MASK, CC_MOS_R_DET_SHIFT, &temp);

	return snprintf(buf, PAGE_SIZE, "%d\n", mos_det_R_table[temp]);
}

static ssize_t threshold2_vdry_store(struct device *dev,
				struct device_attribute *attr,
				const char *buff, size_t size)
{
	int temp = 0;
	unsigned char ret = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	if (kstrtouint(buff, 0, &temp))
		return -EINVAL;

	if (temp >= 0 && temp <= MAX_RESISTANCE) {

		ret = fusb251_resistance_conver_reg(temp);
		if (ret >= 0) {
			fusb251_config_interface(fusb251, FUSB251_THRESHOLD2,
					VDRY_R_DET_MASK, VDRY_R_DET_SHIFT, ret);

			pr_info("[fusb251] set threshold2_vdry to %dk\n", temp);

			return size;
		}
	}

	return -EINVAL;
}

static ssize_t threshold2_vdry_show(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	unsigned char temp = 0;
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	fusb251_read_interface(fusb251, FUSB251_THRESHOLD2,
			VDRY_R_DET_MASK, VDRY_R_DET_SHIFT, &temp);

	return snprintf(buf, PAGE_SIZE, "%d\n", mos_det_R_table[temp]);
}

static ssize_t regdump_show(struct device *dev,
				struct device_attribute *attr,
				char *buf)
{
	int i = 0;
	unsigned char temp = 0;
	char temp_info[200] = "";
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	for (i = 1; i <= FUSB251_REG_NUM; i++) {
		fusb251_read_data(fusb251, i, 1, &temp);
		snprintf(temp_info, PAGE_SIZE, "reg[0x%02x]=0x%02x\n",
				i, temp);
		strcat(buf, temp_info);
	}

	return strlen(buf);
}

DEVICE_ATTR(reset, 0220, NULL, reset_store);
DEVICE_ATTR(control_dry_det, 0660, dry_det_show, dry_det_store);
DEVICE_ATTR(control_sbu_det, 0660, sbu_det_show, sbu_det_store);
DEVICE_ATTR(control_auto_sbu_det, 0660, auto_sbu_det_show,
			auto_sbu_det_store);
DEVICE_ATTR(control_cc_det, 0660, cc_det_show, cc_det_store);
DEVICE_ATTR(status, 0440, status_show, NULL);
DEVICE_ATTR(moi_status, 0440, moi_status_show, NULL);
DEVICE_ATTR(switch_cc, 0660, switch_cc_show, switch_cc_store);
DEVICE_ATTR(switch_sbu, 0660, switch_sbu_show, switch_sbu_store);
DEVICE_ATTR(threshold1_sbu, 0660, threshold1_sbu_show,
			threshold1_sbu_store);
DEVICE_ATTR(threshold1_cc, 0660, threshold1_cc_show,
			threshold1_cc_store);
DEVICE_ATTR(threshold2_vdry, 0660, threshold2_vdry_show,
			threshold2_vdry_store);
DEVICE_ATTR(regdump, 0440, regdump_show, NULL);

static int fusb251_create_attr(struct device *cdev)
{
	int ret = 0;

	ret = device_create_file(cdev, &dev_attr_reset);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_reset\n");
		goto err_attr_reset;
	}

	ret = device_create_file(cdev, &dev_attr_control_dry_det);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_control_dry_det\n");
		goto err_attr_control_dry_det;
	}

	ret = device_create_file(cdev, &dev_attr_control_sbu_det);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_control_sbu_det\n");
		goto err_attr_control_sbu_det;
	}

	ret = device_create_file(cdev, &dev_attr_control_auto_sbu_det);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_control_auto_sbu_det\n");
		goto err_attr_control_auto_sbu_det;
	}

	ret = device_create_file(cdev, &dev_attr_control_cc_det);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_control_cc_det\n");
		goto err_attr_control_cc_det;
	}

	ret = device_create_file(cdev, &dev_attr_status);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_status\n");
		goto err_attr_status;
	}

	ret = device_create_file(cdev, &dev_attr_moi_status);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_moi_status\n");
		goto err_attr_moi_status;
	}

	ret = device_create_file(cdev, &dev_attr_switch_cc);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_switch_cc\n");
		goto err_attr_switch_cc;
	}

	ret = device_create_file(cdev, &dev_attr_switch_sbu);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_switch_sbu\n");
		goto err_attr_switch_sbu;
	}

	ret = device_create_file(cdev, &dev_attr_threshold1_sbu);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_threshold1_sbu\n");
		goto err_attr_threshold1_sbu;
	}

	ret = device_create_file(cdev, &dev_attr_threshold1_cc);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_threshold1_cc\n");
		goto err_attr_threshold1_cc;
	}

	ret = device_create_file(cdev, &dev_attr_threshold2_vdry);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_threshold2_vdry\n");
		goto err_attr_threshold2_vdry;
	}

	ret = device_create_file(cdev, &dev_attr_regdump);
	if (ret < 0) {
		pr_err("[fusb251] failed to create dev_attr_regdump\n");
		goto err_attr_regdump;
	}

	return ret;

err_attr_regdump:
	device_remove_file(cdev, &dev_attr_threshold2_vdry);
err_attr_threshold2_vdry:
	device_remove_file(cdev, &dev_attr_threshold1_cc);
err_attr_threshold1_cc:
	device_remove_file(cdev, &dev_attr_threshold1_sbu);
err_attr_threshold1_sbu:
	device_remove_file(cdev, &dev_attr_switch_sbu);
err_attr_switch_sbu:
	device_remove_file(cdev, &dev_attr_switch_cc);
err_attr_switch_cc:
	device_remove_file(cdev, &dev_attr_moi_status);
err_attr_moi_status:
	device_remove_file(cdev, &dev_attr_status);
err_attr_status:
	device_remove_file(cdev, &dev_attr_control_cc_det);
err_attr_control_cc_det:
	device_remove_file(cdev, &dev_attr_control_auto_sbu_det);
err_attr_control_auto_sbu_det:
	device_remove_file(cdev, &dev_attr_control_sbu_det);
err_attr_control_sbu_det:
	device_remove_file(cdev, &dev_attr_control_dry_det);
err_attr_control_dry_det:
	device_remove_file(cdev, &dev_attr_reset);
err_attr_reset:
	return ret;
}

static void fusb251_delete_attr(struct device *cdev)
{
	device_remove_file(cdev, &dev_attr_regdump);
	device_remove_file(cdev, &dev_attr_threshold2_vdry);
	device_remove_file(cdev, &dev_attr_threshold1_cc);
	device_remove_file(cdev, &dev_attr_threshold1_sbu);
	device_remove_file(cdev, &dev_attr_switch_sbu);
	device_remove_file(cdev, &dev_attr_switch_cc);
	device_remove_file(cdev, &dev_attr_moi_status);
	device_remove_file(cdev, &dev_attr_status);
	device_remove_file(cdev, &dev_attr_control_cc_det);
	device_remove_file(cdev, &dev_attr_control_auto_sbu_det);
	device_remove_file(cdev, &dev_attr_control_sbu_det);
	device_remove_file(cdev, &dev_attr_control_dry_det);
	device_remove_file(cdev, &dev_attr_reset);
}

void fusb251_enable_moisture_detect(bool enable)
{
	if (!g_fusb251) {
		pr_err("[fusb251] enable mos detect: No device is available\n");
		return;
	}

	if (enable) {
		fusb251_i2c_write_reg(g_fusb251, FUSB251_RESET, RESET_MOISTURE);
		mdelay(10);
		fusb251_i2c_write_reg(g_fusb251, FUSB251_CONTROL,
				CTL_EN_CC_DET | CTL_EN_SBU_DET | CTL_EN_DRY_DET);
	} else {
		fusb251_i2c_write_reg(g_fusb251, FUSB251_CONTROL, CTL_DISABLE_ALL);
	}

	pr_info("[fusb251] %s moisture detect\n",
				enable ? "enable" : "disable");
}

static int fusb251_handle_irq(struct fusb251 *fusb251)
{
	unsigned char int_dry_chg;
	unsigned char int_mos_chg;
	unsigned char int_cc1_timer;
	unsigned char int_cc2_timer;
	unsigned char mos_status_fault;
	unsigned char mos_status_sbu1_ft;
	unsigned char mos_status_sbu2_ft;
	unsigned char mos_status_cc1_mos;
	unsigned char mos_status_cc2_mos;
	unsigned char mos_status_sbu1_mos;
	unsigned char mos_status_sbu2_mos;

	if (!fusb251) {
		pr_err("[fusb251] get irq type: No device is available!\n");
		return -1;
	}

	/* Read the interrupt, status and mos_status registers */
	fusb251_read_data(
				fusb251, FUSB251_INTERRUPT, 1,
				&fusb251->registers.interrupt.byte);
	fusb251_read_data(
				fusb251, FUSB251_STATUS, 1,
				&fusb251->registers.status.byte);
	fusb251_read_data(
				fusb251, FUSB251_MOISTATUS, 1,
				&fusb251->registers.mos_status.byte);

	pr_info("[fusb251] int=%2x, st=%2x, mos_st=%2x\n",
				fusb251->registers.interrupt.byte,
				fusb251->registers.status.byte,
				fusb251->registers.mos_status.byte);

	int_dry_chg = fusb251->registers.interrupt.i_dry_chg;
	int_mos_chg = fusb251->registers.interrupt.i_mos_chg;
	int_cc1_timer = fusb251->registers.interrupt.i_cc1_timer;
	int_cc2_timer = fusb251->registers.interrupt.i_cc2_timer;
	mos_status_fault = fusb251->registers.mos_status.fault;
	mos_status_sbu1_ft = fusb251->registers.mos_status.sbu1_ft;
	mos_status_sbu2_ft = fusb251->registers.mos_status.sbu2_ft;
	mos_status_cc1_mos = fusb251->registers.mos_status.cc1_mos;
	mos_status_cc2_mos = fusb251->registers.mos_status.cc2_mos;
	mos_status_sbu1_mos = fusb251->registers.mos_status.sbu1_mos;
	mos_status_sbu2_mos = fusb251->registers.mos_status.sbu2_mos;

	if (int_cc1_timer || int_cc2_timer) {
		if (mos_status_sbu1_ft || mos_status_sbu2_ft) {
			switch_set_state(&fusb251->st_switch, TYPE_MOISTURE);
			pr_info("[fusb251] sbu_ft moisture detected\n");
		} else {
			pr_info("[fusb251] usb attached\n");
		}
	}

	if (int_mos_chg) {
		if (mos_status_cc1_mos || mos_status_cc2_mos) {
			if (mos_status_sbu1_mos || mos_status_sbu2_mos) {
				switch_set_state(&fusb251->st_switch, TYPE_MOISTURE);
				pr_info("[fusb251] cc and sbu moisture detected\n");
			} else {
				if (mos_status_fault) {
					queue_delayed_work(fusb251->fusb251_wq,
						&fusb251->en_mos_det_work, 5 * HZ);
					pr_info("[fusb251] cc_mos=1 but sbu_mos=0\n");
				}
			}
		} else {
			if (int_dry_chg) {
				switch_set_state(&fusb251->st_switch, TYPE_DRY);
				fusb251_enable_moisture_detect(true);
				pr_info("[fusb251] re-enable moisture detect\n");
			}
		}
	}

	return 0;
}

static void fusb251_en_moisture_detect_work(struct work_struct *work)
{
	if (!is_usb_attached)
		fusb251_enable_moisture_detect(true);
}

static void fusb251_eint_work(struct work_struct *work)
{
	struct fusb251 *fusb251 =
		(struct fusb251 *)container_of(work, struct fusb251, eint_work);

	mutex_lock(&fusb251->fusb251_irq_mutex);

	fusb251_handle_irq(fusb251);

	mutex_unlock(&fusb251->fusb251_irq_mutex);
}

static irqreturn_t fusb251_irq_handler(int irq, void *data)
{
	struct fusb251 *fusb251 = data;

	queue_work(fusb251->fusb251_wq, &fusb251->eint_work);

	return IRQ_HANDLED;
}

static void fusb251_init(struct fusb251 *fusb251)
{
	/* reset moisture */
	fusb251_i2c_write_reg(fusb251, FUSB251_RESET, RESET_MOISTURE);
	/* mask all the interrupt */
	fusb251_i2c_write_reg(fusb251, FUSB251_INT_MASK, MSK_I_MASK_ALL);
	/*
	 * CC moisture detection time
	 * 400us settle time with 3 time adc reading
	 */
	fusb251_i2c_write_reg(fusb251, FUSB251_TIMER2,
			ADC_READ_TIMES_3 | CC_SETTLE_400US);
	/* cc and sbu moisture detection threshold ,default set to 480kohm */
	fusb251_i2c_write_reg(fusb251, FUSB251_THRESHOLD1,
			SBU_MOS_R_DET_480K | CC_MOS_R_DET_480K);
	/*
	 * sbu float detection and dry check threshold ,
	 * default set to sbu_float_det:200mv and vdry:480kohm
	 */
	fusb251_i2c_write_reg(fusb251, FUSB251_THRESHOLD2,
			SBU_FLOAT_DET_200MV | VDRY_DET_480K);
	/* dry check cycle time ,default set to 10s */
	fusb251_i2c_write_reg(fusb251, FUSB251_TIMER1, TIMER_DRY_DET_10S);
	/* mask ovp interrupt */
	fusb251_i2c_write_reg(fusb251, FUSB251_INT_MASK,
			MSK_I_OVP | MSK_I_OVP_REC);
}

static int fusb251_request_irq(struct fusb251 *fusb251)
{
	int ret = 0;
	struct device_node *np = fusb251->dev->of_node;

	fusb251->moisture_en = of_get_named_gpio(np, "moisture_det_en", 0);
	if (fusb251->moisture_en < 0) {
		pr_err("[fusb251] unable to get irq gpio\n");
		return -1;
	}

	if (!gpio_is_valid(fusb251->moisture_en)) {
		pr_err("[fusb251] irq gpio is invalid\n");
		return -1;
	}

	ret = gpio_request(fusb251->moisture_en, "moisture_det_en");
	if (ret) {
		pr_err("[fusb251] irq gpio request failed\n");
		return -1;
	}

	ret = gpio_direction_input(fusb251->moisture_en);
	if (ret) {
		pr_err("[fusb251] set_direction for irq gpio failed\n");
		goto err_irq_gpio_dir;
	}

	fusb251->irq = gpio_to_irq(fusb251->moisture_en);
	ret = request_irq(fusb251->irq, fusb251_irq_handler,
		IRQF_TRIGGER_FALLING | IRQF_ONESHOT, "FUSB251_MOS_EN", fusb251);
	if (ret < 0) {
		pr_err("[fusb251] request thread irq failed\n");
		goto err_irq_request;
	}

	return 0;

err_irq_request:
err_irq_gpio_dir:
	if (gpio_is_valid(fusb251->moisture_en))
		gpio_free(fusb251->moisture_en);
	return -1;
}

static int fusb251_i2c_probe(struct i2c_client *client,
			const struct i2c_device_id *id)
{
	int ret = 0;
	struct fusb251 *fusb251 = NULL;

	if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C)) {
		pr_err("[fusb251] check_functionality failed\n");
		return -ENODEV;
	}

	fusb251 = kzalloc(sizeof(*fusb251), GFP_KERNEL);
	if (!fusb251)
		return -ENOMEM;

	memset(fusb251, 0, sizeof(*fusb251));

	fusb251->i2c = client;
	fusb251->dev = &client->dev;
	i2c_set_clientdata(client, fusb251);

	fusb251->st_switch.name = "ld";
	fusb251->st_switch.index = 0;
	fusb251->st_switch.state = TYPE_DRY;
	ret = switch_dev_register(&fusb251->st_switch);
	if (ret) {
		pr_err("[fusb251] switch_dev_register failed\n");
		goto err_switch;
	}

	g_fusb251 = fusb251;

	sema_init(&fusb251->suspend_lock, 1);
	mutex_init(&fusb251->fusb251_irq_mutex);
	mutex_init(&fusb251->fusb251_i2c_mutex);

	fusb251->fusb251_wq = create_singlethread_workqueue("fusb251_wq");
	if (!fusb251->fusb251_wq) {
		pr_err("[fusb251] create workqueue failed\n");
		goto err_wq;
	}

	INIT_WORK(&fusb251->eint_work, fusb251_eint_work);
	INIT_DELAYED_WORK(&fusb251->en_mos_det_work,
		fusb251_en_moisture_detect_work);

	fusb251_init(fusb251);

	ret = fusb251_request_irq(fusb251);
	if (ret < 0)
		goto err_irq;

	ret = fusb251_create_attr(fusb251->dev);
	if (ret < 0)
		goto err_attr;

	pr_info("[fusb251] probe successfully\n");

	return 0;

err_attr:
	free_irq(fusb251->irq, fusb251);
	if (gpio_is_valid(fusb251->moisture_en))
		gpio_free(fusb251->moisture_en);
err_irq:
	if (fusb251->fusb251_wq)
		destroy_workqueue(fusb251->fusb251_wq);
err_wq:
	switch_dev_unregister(&fusb251->st_switch);
err_switch:
	kfree(fusb251);
	fusb251 = NULL;

	return ret;
}

static int fusb251_i2c_remove(struct i2c_client *client)
{
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	free_irq(fusb251->irq, fusb251);

	if (gpio_is_valid(fusb251->moisture_en))
		gpio_free(fusb251->moisture_en);

	fusb251_delete_attr(fusb251->dev);

	if (fusb251->fusb251_wq)
		destroy_workqueue(fusb251->fusb251_wq);

	i2c_unregister_device(client);

	kfree(fusb251);
	fusb251 = NULL;

	return 0;
}

static int fusb251_i2c_resume(struct device *dev)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	pr_err("[fusb251] %s enter\n", __func__);

	if (!fusb251) {
		pr_err("[fusb251] resume: No device is available!\n");
		return -EINVAL;
	}

	up(&fusb251->suspend_lock);

	return 0;
}

static int fusb251_i2c_suspend(struct device *dev)
{
	struct i2c_client *client = to_i2c_client(dev);
	struct fusb251 *fusb251 =
			(struct fusb251 *)i2c_get_clientdata(client);

	pr_err("[fusb251] %s enter\n", __func__);

	if (!fusb251) {
		pr_err("[fusb251] suspend: No device is available!\n");
		return -EINVAL;
	}

	down(&fusb251->suspend_lock);

	return 0;
}

static const struct i2c_device_id fusb251_i2c_id[] = {
	{"fusb251", 0},
	{},
};

#ifdef CONFIG_OF
static const struct of_device_id of_fusb251_match_table[] = {
	{.compatible = "on,fusb251",},
	{},
};
MODULE_DEVICE_TABLE(of, of_fusb251_match_table);
#endif

#ifdef CONFIG_PM
static const struct dev_pm_ops fusb251_dev_pm_ops = {
	.suspend = fusb251_i2c_suspend,
	.resume  = fusb251_i2c_resume,
};
#endif

static struct i2c_driver fusb251_i2c_driver = {
	.driver = {
		.name = "fusb251",
		.owner = THIS_MODULE,
		.of_match_table = of_match_ptr(of_fusb251_match_table),
#ifdef CONFIG_PM
		.pm = &fusb251_dev_pm_ops,
#endif
	},
	.probe =    fusb251_i2c_probe,
	.remove =   fusb251_i2c_remove,
	.id_table = fusb251_i2c_id,
};

static int __init fusb251_i2c_init(void)
{
	int ret = 0;

	ret = i2c_add_driver(&fusb251_i2c_driver);
	if (ret != 0)
		pr_err("[fusb251] fusb251 i2c init failed!\n");

	return ret;
}
module_init(fusb251_i2c_init);

static void __exit fusb251_i2c_exit(void)
{
	i2c_del_driver(&fusb251_i2c_driver);
}
module_exit(fusb251_i2c_exit);

MODULE_AUTHOR("jianli.gao@onsemi.com");
MODULE_DESCRIPTION("TYPEC FUSB251 driver");
MODULE_LICENSE("GPL v2");
